#!/usr/bin/env python3

import babeltrace
from enum import IntFlag

defaultTracePath = "./trace-ctf-nanos6/ust/uid/1042/64-bit/"

class TaskLabelsDB:
	__db = {}
	def addLabel(self, label, taskTypeID):
		if (taskTypeID in self.__db) and (self.__db[taskTypeID] != label):
			raise Exception("Attempt to add another different label for task ID: " + str(taskTypeID) + " old label = " + self.__db[taskTypeID] + " new label = " + label)
		self.__db[taskTypeID] = label
	def getLabel(self, taskTypeID):
		if not taskTypeID in self.__db:
			raise Exception("task with ID: " + str(taskTypeID) + " is not in dictionary")
		return self.__db[taskTypeID]
	def checkConsistency(self):
		allIDs = self.__db.keys().sort()
		cnt = 0
		for taskTypeID in allIDs:
			if taskTypeID != cnt:
				raise Exception("task with type id " + str(taskTypeID) + " is missing")
			cnt += 1

class TaskStatus(IntFlag):
	UNINITIALIZED = 2
	ADDED         = 4
	RUNNING       = 8
	FINISHED      = 16
	BLOCKED       = 32

class Task:
	taskLabelsDB = None
	__statusTransitions = {
		TaskStatus.UNINITIALIZED : TaskStatus.ADDED,
		TaskStatus.ADDED         : TaskStatus.RUNNING,
		TaskStatus.RUNNING       : TaskStatus.BLOCKED | TaskStatus.FINISHED,
		TaskStatus.BLOCKED       : TaskStatus.RUNNING,
		TaskStatus.FINISHED      : 0
	}

	def __init__(self):
		self.__taskID     = 0
		self.__taskTypeID = 0
		self.__taskLabel  = ""
		self.__status     = TaskStatus.UNINITIALIZED

	def init(self, taskTypeID, taskID):
		self.__taskTypeID = taskTypeID
		self.__taskID     = taskID
		self.__taskLabel  = Task.taskLabelsDB.getLabel(taskTypeID)
		self.changeStatus(TaskStatus.ADDED)

	def getStatus(self):
		return self.__status

	def changeStatus(self, new):
		accepted = self.__statusTransitions[self.__status]
		if not (accepted & new):
			raise Exception("Task ID " + str(task.taskID) + " attempted to change from status " + TaskStatus(old).name + " to " + TaskStatus(new).name)
		self.__status = new

	def getTaskID(self):
		return self.__taskID

class TaskProvider():
	def __init__(self):
		self.usingTaskDict = {}
		self.disposedTaskList = []

	def setup(self, numTasks = 0):
            pass

	def getNewTask(self):
		raise Exception("Not Implemented!")

	def validate(self):
		raise Exception("Not Implemented!")

	def validateCommon(self):
		if self.usingTaskDict:
			print("Warning: " + str(len(self.usingTaskDict)) + " tasks are still being used")

		for task in self.disposedTaskList:
			if task.getStatus() != TaskStatus.FINISHED:
				raise Exception("Error: Task " + str(task.taskID) + " has not finished")

	def addTask(self, task):
		self.usingTaskDict[task.getTaskID()] = task

	def getTask(self, taskID):
		return self.usingTaskDict[taskID]

	def disposeTask(self, task):
		assert(task != None)
		disposable = self.usingTaskDict.pop(task.getTaskID(), None)
		assert(task == disposable)
		self.disposedTaskList.append(task)

class TaskProviderUnlimited(TaskProvider):
	def __init__(self):
		super().__init__()

	def getNewTask(self):
		task = Task()
		return task

	def validate(self):
		self.validateCommon()

class TaskProviderPool(TaskProvider):
	def __init__(self):
		super().__init__()
		self.__freeTaskList = []
		self.__numTasks = 0

	def setup(self, numTasks):
		self.__freeTaskList = [Task() for _ in range(numTasks)]
		self.__numTasks = numTasks

	def getNewTask(self):
		try:
			task = self.__freeTaskList.pop()
		except:
			print("Error: No more free tasks left")
			raise

		return task

	def validate(self):
		self.validateCommon()

		if self.__freeTaskList:
			print("Warning: " + str(len(self.__freeTaskList)) + " tasks have not been used")

		if len(self.disposedTaskList) != self.__numTasks:
			print("self.__numTasks ", self.__numTasks)
			print("len of disposedTaskList ", len(self.disposedTaskList))
			print("Warning: " + str(self.__numTasks - len(self.disposedTaskList)) + " tasks have not been disposed")





def loadTrace(name = defaultTracePath):
	col = babeltrace.TraceCollection()
	if col.add_trace(name, 'ctf') is None:
		raise RuntimeError('Cannot add trace')

	return col

def checkTasks(trace, numTasks = 0):
	cnt = 0
	taskLabelsDB = TaskLabelsDB()
	Task.taskLabelsDB = taskLabelsDB
	taskProvider = None
	if numTasks:
		taskProvider = TaskProviderPool()
	else:
		taskProvider = TaskProviderUnlimited()
	taskProvider.setup(numTasks)

	print(" === reading events ===")
	for event in trace.events:
		print(event.name, event["cpu_id"])
		if event.name == "nanos6:task_label":
			label      = event["label"]
			taskTypeID = event["type"]
			taskLabelsDB.addLabel(label, taskTypeID)
		elif event.name == "nanos6:task_add":
			taskTypeID = event["type"]
			taskID     = event["id"]
			task = taskProvider.getNewTask()
			task.init(taskTypeID, taskID)
			taskProvider.addTask(task)
		elif event.name == "nanos6:task_execute":
			taskID = event["id"]
			task = taskProvider.getTask(taskID)
			task.changeStatus(TaskStatus.RUNNING)
			pass
		elif event.name == "nanos6:task_block":
			taskID = event["id"]
			task = taskProvider.getTask(taskID)
			task.changeStatus(TaskStatus.BLOCKED)
			pass
		elif event.name == "nanos6:task_end":
			taskID = event["id"]
			task = taskProvider.getTask(taskID)
			task.changeStatus(TaskStatus.FINISHED)
			taskProvider.disposeTask(task)
			pass
		else:
			raise Exception("unknown CTF event: " + event.name)

		cnt += 1

	print(" === validating " + str(cnt) + " events ===")
	taskProvider.validate()
	del taskProvider

if __name__ == '__main__':
	print("Hey! load me as a module!")
